/**
 * Copyright (c) 2021 Guido Kramann
 *
 * This file is part of a project called kaleidophone - http://www.kramann.info/kaleidophone
 *
 * ---------------------------------------------------------------------------------------------------------
 * Bild is a german word and it means image
 *
 * This file provides methods to draw circles and to project a 3D space with spheres in a plane.
 * 
 * Coordinates of circle for instance are mapped from -1..+1 in both dimensions to the image size in pixels.
 * ---------------------------------------------------------------------------------------------------------
 *
 * kaleidophone is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * kaleidophone is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the kaleidophone. If not, see <http://www.gnu.org/licenses/>.
*/

function Bild()
{
    //int[] m; //represents the image
    var m;
    //int[] n; //same size as m. Used for buffering and pre-calculation.
    var n;
    var width,height;
    
    // color-codes
    let COLOR = [
        [0,0,0],     // 0 black
        [0,0,255],   // 1 blue    represents the prime number 2
        [255,0,0],   // 2 red     represents the prime number 3   
        [0,255,0],   // 3 green   represents the prime number 5
        [255,255,0], // 4 yellow  represents the prime number 7
        [127,127,127]];// 5 grey    multiple grey fields represent an operator
    
    
    
    function makeArrayZerosZeros(zeilen, spalten)
    {
         var arr = new Array(zeilen);
         for(var i=0;i<arr.length;i++)
         {
               arr[i] = new Array(spalten);
         }
         for(var i=0;i<arr.length;i++)
            for(var k=0;k<arr[i].length;k++)
                arr[i][k] = 0;


         return arr;
    }

    function makeZeros(zeilen)
    {
         var arr = new Array(zeilen);
         for(var i=0;i<arr.length;i++)
         {
               arr[i] = 0;
         }
         return arr;
    }
    
    this.init = function(width_and_height)
    {
        width = width_and_height;
        height = width_and_height;
        m = makeZeros(width*height);
        n = makeZeros(width*height);
    }
    
    this.clear = function()
    {
         for(var i=0;i<m.length;i++) m[i]=0;
    }

    this.clear = function(q)
    {
         for(var i=0;i<q.length;i++) q[i]=0;
    }
    
    function xpix2x(x){ return  -1.0 + 2.0 * x / width;}
    function ypix2y(y){ return   1.0 - 2.0 * y / height;}
    function x2xpix(x){ return  Math.floor(0.5*(x+1.0)*width);}
    function y2ypix(y){ return  Math.floor(0.5*(-y+1.0)*height);}
    function inx2x(inx){ return  inx%width;}
    function inx2y(inx){ return  Math.floor(inx/width);}
    
    this.circle = function(x,y,radius,colcode)
    {
        for(var i=0;i<m.length;i++)
        {
            var x2 = xpix2x(inx2x(i));
            var y2 = ypix2y(inx2y(i));
            var dist = Math.sqrt((x-x2)*(x-x2)+(y-y2)*(y-y2));
            if(dist<=radius) m[i]=colcode;
        }
    }
    
    this.draw = function(qapplet,xoff,yoff,breite,hoehe)
    {
         var dx = breite/width;
         var dy = hoehe/height;
         qapplet.noStroke();
         for(var i=0;i<m.length;i++)
         {
              var c = m[i];
              var x = inx2x(i);
              var y = inx2y(i);
              qapplet.fill(COLOR[c][0],COLOR[c][1],COLOR[c][2]);
              qapplet.rect(xoff+x*dx,yoff+y*dy,dx,dy);
         }
    }
}
