/**
 * Copyright (c) 2021 Guido Kramann
 *
 * This file is part of a project called kaleidophone - http://www.kramann.info/kaleidophone
 *
 * ------------------------------------------------------------------------------------------------------
 * This file allows to make generate software based musical instruments on the base of physical modeling.
 * ------------------------------------------------------------------------------------------------------
 *
 * kaleidophone is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * kaleidophone is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the kaleidophone. If not, see <http://www.gnu.org/licenses/>.
*/

function Instrument()
{
    //Kopplung: 0->1 0->2    1->0 1->2   2->0 2->1
    var dt_kopplung=null;
    var amp_kopplung=null;
    var arr_kopplung=null;  // innere Arrays!
    var index_kopplung = 0;

    var dt1_reverb = 0.005;
    var dt2_reverb = 0.8;
    var reverbL = new Array();
    var reverbR = new Array();
    var inx_reverb=0;
    var aLL = 0.4;
    var aRR = 1.0 - aLL;
    var LEFT=0;
    var RIGHT=0;
    var LLEFT=null;
    var RRIGHT=null;

    var midimin=0,midimax=0;
    var faktor_ot1=0,faktor_ot2=0,faktor_ot3=0;
    var amp_ot1=0,amp_ot2=0,amp_ot3=0;
    var Qmin=0, Qmax=0;
    var dt_F=0, F0=0,F1=0,F2=0;

    var obertonA=null;
    var obertonB=null;
    var obertonC=null;

    var t=0;

    var Fplay=null; //Kraftfaktor mit dem gespielt wird
    var Tplay=null; //Zeitpunkt zu dem begonnen wird zu spielen
    var Pplay=null; //Panning jeder Tonhoehe
    var Aplay=null; //Panning jeder Tonhoehe

    this.init = function(xmidimin,xmidimax,
                         xfaktor_ot1,xfaktor_ot2,xfaktor_ot3,
                         xamp_ot1,xamp_ot2,xamp_ot3,
                         xQmin, xQmax,
                         xdt_F, xF0,xF1,xF2,
                         dtk01, dtk02, dtk10, dtk12, dtk20, dtk21,  //Kopplung Tt
                         apk01, apk02, apk10, apk12, apk20, apk21,  //Kopplung Verst. 
                         PAN)
    {
        index_kopplung = 0;         
        dt_kopplung = new Array(dtk01, dtk02, dtk10, dtk12, dtk20, dtk21);
        amp_kopplung = new Array(apk01, apk02, apk10, apk12, apk20, apk21);
        arr_kopplung = new Array(6);
        for(var i=0;i<arr_kopplung.length;i++)
        {
            var anz = Math.ceil(dt_kopplung[i] / dt);  //geteilt durch Schritte peo Sample
            arr_kopplung[i] = new Array(anz);
            for(var k=0;k<arr_kopplung[i].length;k++)
                arr_kopplung[i][k] = 0.0;
        }

        var anz1 = Math.round(sample_rate*dt1_reverb);
        var anz2 = Math.round(dt2_reverb / dt1_reverb);
        reverbL = new Array(anz2);
        reverbR = new Array(anz2);
        for(var i=0;i<reverbL.length;i++)
        {
            reverbL[i] = new Array(anz1);
            reverbR[i] = new Array(anz1);

            for(var k=0;k<reverbL[i].length;k++)
            {
                reverbL[i][k] = 0.0;
                reverbR[i][k] = 0.0;
            }
        }
        //alert("reverb.length="+reverb.length+" reverb[0].length="+reverb[0].length);

        aLL = PAN;
        aRR = 1.0 - PAN;

        midimin = xmidimin;
        midimax = xmidimax;
        faktor_ot1 = xfaktor_ot1;
        faktor_ot2 = xfaktor_ot2;
        faktor_ot3 = xfaktor_ot3;
        amp_ot1 = xamp_ot1;
        amp_ot2 = xamp_ot2;
        amp_ot3 = xamp_ot3;
        Qmin = xQmin;
        Qmax = xQmax;
        dt_F = xdt_F;
        F0 = xF0;
        F1 = xF1;
        F2 = xF2;

         t = 0.0;

        Fplay = new Array(midimax-midimin+1);  //Kraft
        Tplay = new Array(midimax-midimin+1);  //Startzeit
        Pplay = new Array(midimax-midimin+1);  //Panning
        Aplay = new Array(midimax-midimin+1);  //Tiefere weniger

        LLEFT = new Array(midimax-midimin+1);
        RRIGHT = new Array(midimax-midimin+1);

        for(var i=0;i<LLEFT.length;i++)
        {
            LLEFT[i] = 0.0;
            RRIGHT[i] = 0.0;
        }

        for(var i=0;i<Fplay.length;i++)
        {
            Fplay[i] = 0.0;
            Tplay[i] = -9999999.0;
            Pplay[i] = 0.2*(i/(Pplay.length-1)) - 0.1;

            //Tiefere weniger stark:
            Aplay[i] = (0.1 + (i/(Pplay.length-1)))/1.1;
        }

        obertonA = new Array(midimax-midimin+1);
        obertonB = new Array(midimax-midimin+1);
        obertonC = new Array(midimax-midimin+1);
        for(var i=0;i<obertonA.length;i++)
        {
            obertonA[i] = new Oberton();
            obertonA[i].init(dt,faktor_ot1*midi2Frequency(midimin+i),(Qmax+Qmin)/2,amp_ot1); // Dt,Freq,Q,Amp
            obertonB[i] = new Oberton();
            obertonB[i].init(dt,faktor_ot2*midi2Frequency(midimin+i),(Qmax+Qmin)/2,amp_ot2); // Dt,Freq,Q,Amp
            obertonC[i] = new Oberton();
            obertonC[i].init(dt,faktor_ot3*midi2Frequency(midimin+i),(Qmax+Qmin)/2,amp_ot3); // Dt,Freq,Q,Amp
        }
    }

    this.calculateEnergy = function(midi)
    {
        var energy = 0.0;
        energy += obertonA[midi-midimin].calculateEnergy();
        energy += obertonB[midi-midimin].calculateEnergy();
        energy += obertonC[midi-midimin].calculateEnergy();
        return energy;
    }

    this.step = function()
    {
        LEFT = 0.0;
        RIGHT = 0.0;

        for(var i=0;i<Tplay.length;i++)
        {
            var aL = aLL+Pplay[i]; 
            var aR = aRR-Pplay[i]; 

            var F = 0.0;
            var x = 0.0;
            var y = 0.0;
            if(t-Tplay[i]<dt_F)
            {
                F = F0 + (F1-F0)*(t-Tplay[i])/dt;
                F*=Fplay[i];
            }
            else if(t-Tplay[i]<2.0*dt_F)
            {
                F = F1 + (F2-F1)*(t-Tplay[i]-dt)/dt;    
                F*=Fplay[i];
            }

            index_kopplung++;

            var FA = F;// + arr_kopplung[2][index_kopplung%arr_kopplung[2].length] + arr_kopplung[4][index_kopplung%arr_kopplung[4].length];
            var FB = F;// + arr_kopplung[0][index_kopplung%arr_kopplung[0].length] + arr_kopplung[5][index_kopplung%arr_kopplung[5].length];
            var FC = F;// + arr_kopplung[1][index_kopplung%arr_kopplung[1].length] + arr_kopplung[3][index_kopplung%arr_kopplung[3].length];

            var xA = obertonA[i].step(FA);
            var xB = obertonB[i].step(FB);
            var xC = obertonC[i].step(FC);
/*
            arr_kopplung[0][index_kopplung%arr_kopplung[0].length] = xA*amp_kopplung[0];  // 0->1
            arr_kopplung[1][index_kopplung%arr_kopplung[1].length] = xA*amp_kopplung[1];  // 0->2
            arr_kopplung[2][index_kopplung%arr_kopplung[2].length] = xB*amp_kopplung[2];  // 1->0
            arr_kopplung[3][index_kopplung%arr_kopplung[3].length] = xB*amp_kopplung[3];  // 1->2
            arr_kopplung[4][index_kopplung%arr_kopplung[4].length] = xC*amp_kopplung[4];  // 2->0
            arr_kopplung[5][index_kopplung%arr_kopplung[5].length] = xC*amp_kopplung[5];  // 2->1
*/
            x += xA+xB+xC;

            y=x;
            x*=aL;
            y*=aR;
/*
            for(var k=1;k<reverb.length;k++)
            {
                x += aR*0.57*reverb[k][inx_reverb];  //neuster von k-1 ist ältester von k
                y += aL*0.57*reverb[k][inx_reverb];  //neuster von k-1 ist ältester von k
            }

            var inx2 = (inx_reverb+1)%reverb[0].length;
            reverb[0][inx_reverb] = x;

            for(var k=1;k<reverb.length;k++)
            {
                reverb[k][inx_reverb] = 0.57*reverb[k-1][inx2];  //neuster von k-1 ist ältester von k
            }

            inx_reverb++;
            inx_reverb%=reverb[0].length;
*/
            LEFT+=x*Aplay[i];
            RIGHT+=y*Aplay[i];
            LLEFT[i]=x*Aplay[i];
            RRIGHT[i]=y*Aplay[i];
        }
/*
            reverbL[0][inx_reverb] = LEFT;
            reverbR[0][inx_reverb] = RIGHT;
            for(var k=1;k<reverbL.length;k++)
            {
                LEFT += aR*reverbL[k][inx_reverb] + 0.4*aL*reverbR[k][inx_reverb];  //neuster von k-1 ist ältester von k
                RIGHT += aL*reverbR[k][inx_reverb] + 0.4*aR*reverbL[k][inx_reverb];  //neuster von k-1 ist ältester von k
            }

            var inx2 = (inx_reverb+1)%reverbL[0].length;

            for(var k=1;k<reverbL.length;k++)
            {
                reverbL[k][inx_reverb] = 0.8*reverbL[k-1][inx2];  //neuster von k-1 ist ältester von k
                reverbR[k][inx_reverb] = 0.8*reverbR[k-1][inx2];  //neuster von k-1 ist ältester von k
            }

            inx_reverb++;
            inx_reverb%=reverbL[0].length;
*/

        t+=dt;

    }   



    this.getLEFT = function()
    {
       return LEFT;
    }

    this.getRIGHT = function()
    {
       return RIGHT;
    }

    this.getLLEFT = function(midi)
    {
       return LLEFT[midi-midimin];
    }

    this.getRRIGHT = function(midi)
    {
       return RRIGHT[midi-midimin];
    }

    this.play = function(midi,velocity)
    {
        Tplay[midi-midimin] = t;
        Fplay[midi-midimin] = velocity;
    }

    this.damp = function(midi)
    {
        if(midi - midimin>=0 && midi - midimin<obertonA.length)
        {
            obertonA[midi - midimin].setVtoZero();
            obertonB[midi - midimin].setVtoZero();
            obertonC[midi - midimin].setVtoZero();
            //obertonA[midi - midimin].init(dt,faktor_ot1*midi2Frequency(midi),Qmax,amp_ot1); // Dt,Freq,Q,Amp
            //obertonB[midi - midimin].init(dt,faktor_ot2*midi2Frequency(midi),Qmax,amp_ot2); // Dt,Freq,Q,Amp
            //obertonC[midi - midimin].init(dt,faktor_ot3*midi2Frequency(midi),Qmax,amp_ot3); // Dt,Freq,Q,Amp
            obertonA[midi - midimin].updateCD(5.0*Qmax); // Dt,Freq,Q,Amp
            obertonB[midi - midimin].updateCD(5.0*Qmax); // Dt,Freq,Q,Amp
            obertonC[midi - midimin].updateCD(5.0*Qmax); // Dt,Freq,Q,Amp
            //Tplay[midi-midimin] = 0;
            //Fplay[midi-midimin] = 0;
        }
    }

    this.undamp = function(midi)
    {
            obertonA[midi - midimin].updateCD((Qmax+Qmin)/2); // Dt,Freq,Q,Amp
            obertonB[midi - midimin].updateCD((Qmax+Qmin)/2); // Dt,Freq,Q,Amp
            obertonC[midi - midimin].updateCD((Qmax+Qmin)/2); // Dt,Freq,Q,Amp
    }

    this.stop = function(midi)
    {
    }

    this.getMidimin = function()
    {
        return midimin;
    }

    this.getMidimax = function()
    {
        return midimax;
    }
    this.reset = function()
    {
        t = 0.0;
        for(var i=0;i<obertonA.length;i++)
        {
            obertonA[i].reset();
            obertonB[i].reset();
            obertonC[i].reset();
        }
        for(var i=0;i<Fplay.length;i++)
        {
            Fplay[i] = 0.0;
            Tplay[i] = -9999999.0;
        }
    }

}



