/**
 * Copyright (c) 2021 Guido Kramann
 *
 * This file is part of a project called kaleidophone - http://www.kramann.info/kaleidophone
 *
 *
 * kaleidophone is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * kaleidophone is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the kaleidophone. If not, see <http://www.gnu.org/licenses/>.
*/

function InstrumentWav()
{
    var DAMP = 4000;

    var LEFT=0;
    var RIGHT=0;

    var sample;
    var leftakt;
    var rightakt;
    var inxakt;
    var dampinxakt;
    var velakt;
    var midiakt;
    var midimin;
    var midimax;
    this.init = function(xsample,maxpoly,xmidimin,xmidimax)
    {
        sample = xsample;
        midimin = xmidimin;
        midimax = xmidimax;
        leftakt = new Array(maxpoly);
        rightakt = new Array(maxpoly);
        inxakt = new Array(maxpoly);
        dampinxakt = new Array(maxpoly);
        velakt = new Array(maxpoly);
        midiakt = new Array(maxpoly);
        for(var i=0;i<leftakt.length;i++)
        {
            leftakt[i] = null;
            rightakt[i] = null;
            inxakt[i]=0;
            dampinxakt[i]=0;
            velakt[i]=0.0;
            midiakt[i]=0;
        }        
    }

    this.play = function(midi,velocity)
    {
        for(var i=0;i<leftakt.length;i++)
        {
            if(leftakt[i]==null)
            {
                var inx = (midi-midimin)*2;
                leftakt[i] = sample[inx];
                rightakt[i] = sample[inx+1];
                inxakt[i] = 0;
                dampinxakt[i] = 0;
                velakt[i] = velocity;
                midiakt[i] = midi;
                return;
            }
        }
    }

    this.getLEFT = function()
    {
       return LEFT;
    }

    this.getRIGHT = function()
    {
       return RIGHT;
    }

    this.step = function()
    {
        LEFT=0;
        RIGHT=0;

        for(var i=0;i<leftakt.length;i++)
        {
            if(leftakt[i]!=null && dampinxakt[i]>0 && inxakt[i]>=dampinxakt[i]+DAMP)
            {
                leftakt[i] = null;
                rightakt[i] = null;
                inxakt[i] = 0;
                dampinxakt[i] =0;
            }
            else if(leftakt[i]!=null && inxakt[i]<leftakt[i].length)
            {
                var fkt = 1.0;
                if(dampinxakt[i]>0)
                    fkt = 1.0 - (inxakt[i]-dampinxakt[i])/DAMP;
                LEFT+=leftakt[i][inxakt[i]]*velakt[i]*fkt;
                RIGHT+=rightakt[i][inxakt[i]]*velakt[i]*fkt;
//                LEFT+=leftakt[i][inxakt[i]];
//                RIGHT+=rightakt[i][inxakt[i]];
                inxakt[i]++;
            }
            else if(leftakt[i]!=null)
            {
                leftakt[i] = null;
                rightakt[i] = null;
                inxakt[i] = 0;
                dampinxakt[i] =0;
            }
        }
    }   

    //alle anderen Methoden als Dummy-Methoden:


    this.calculateEnergy = function(midi)
    {
        return 0.0;
    }

    this.getLLEFT = function(midi)
    {
       return 0.0;
    }

    this.getRRIGHT = function(midi)
    {
       return 0.0;
    }

    this.damp = function(midi)
    {
        for(var i=0;i<leftakt.length;i++)
        {
            if(leftakt[i]!=null && midiakt[i]==midi && inxakt[i]<leftakt[i].length-DAMP*2)
            {
                dampinxakt[i]=inxakt[i];
            }
        }
    }

    this.undamp = function(midi)
    {
        for(var i=0;i<leftakt.length;i++)
        {
            if(leftakt[i]!=null && midiakt[i]==midi)
            {
                dampinxakt[i]=0;
            }
        }
    }

    this.stop = function(midi)
    {
    }

    this.getMidimin = function()
    {
        return midimin;
    }

    this.getMidimax = function()
    {
        return midimax;
    }
    this.reset = function()
    {
        for(var i=0;i<leftakt.length;i++)
        {
            leftakt[i] = null;
            rightakt[i] = null;
            inxakt[i] = 0;
            velakt[i] = 0.0;
        }
    }
 
}

