/**
 * Copyright (c) 2021 Guido Kramann
 *
 * This file is part of a project called kaleidophone - http://www.kramann.info/kaleidophone
 *
 * ---------------------------------------------------------------------------------------------------------
 * Manager is able to collect midi events and to decide, which ones from the passed should be dampened.
 *
 * ---------------------------------------------------------------------------------------------------------
 *
 * kaleidophone is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * kaleidophone is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the kaleidophone. If not, see <http://www.gnu.org/licenses/>.
*/


function Manager()
{
    var event; //consists of other arrays in which is...
    //Index 0          1        2         3          4      5   6    7    8
    //      id of ev.  time_ms  tick      inst_nr    midi   p   q    r    s  means: freq= 2^p*3^q*5^r*7^s*FACTOR
    // consequently any event should be damped if Basenumber changes, that is: a new harmonic context appears(?)  
    var ID = 0;
    var INDEX  = 0;
    this.init = function()
    {
        event = new Array(100); 
        ID=0;
        INDEX = 0;
    }   

    this.addEvent = function(tick, inst_nr, midi,p,q,r,s)
    {
        if(LOGGING) console.log("add "+midi);

        event[INDEX] = new Array(ID,Date.now(),tick,inst_nr,midi,p,q,r,s);
        ID++;
        INDEX++;
        INDEX%=event.length;
    }    

    this.returnAndRemoveTooOld = function(time_horizon_ms,tick,tick_modulo)
    {
        var t = Date.now();
        for(var i=0;i<event.length;i++)
        {
             var ev = event[i];
             if(ev!=null && ev[1]+time_horizon_ms<t && tick%tick_modulo==0) //nur auf ganzen Takten zu lange beenden
             {
                 event[i] = null;
                 return ev;
             }
        }
  
        return null;
    }

    this.returnAndRemoveOldest = function()
    {
        var t = Date.now();
        var inx = -1;
        for(var i=0;i<event.length;i++)
        {
            if(event[i]!=null && event[i][1]<=t)
            {
                 inx = i;
                 t = event[i][1];
            }
        }

        if(inx>-1)
        {
            if(LOGGING) console.log("remove "+event[inx][4]);
            var ev = event[inx];
            event[inx] = null;
            return ev;
        }
        else
        {
            return null;
        }
    }

    this.degreeOfDissonance = function(p,q,r,s,  new_p,new_q,new_r,new_s)
    {
        return 1*iabs(p - new_p) + 2*iabs(q - new_q) + 4*iabs(r - new_r) + 6*iabs(s - new_s);
    }

    this.degreeOfDissonancePot = function(pot1,pot2)
    {
        var p = pot1[0];
        var q = pot1[1];
        var r = pot1[2];
        var s = pot1[3];

        var new_p = pot2[0];
        var new_q = pot2[1];
        var new_r = pot2[2];
        var new_s = pot2[3];


        return 1*iabs(p - new_p) + 2*iabs(q - new_q) + 4*iabs(r - new_r) + 6*iabs(s - new_s);
    }

    this.returnAndRemoveTooDissonant = function(MAX_DISSO,   new_p,new_q,new_r,new_s)
    {
        for(var i=0;i<event.length;i++)
        {
          if(event[i]!=null)
          {
            var ev = event[i];
            var p = ev[5];
            var q = ev[6];
            var r = ev[7];
            var s = ev[8];

            var gradus = 1*iabs(p - new_p) + 2*iabs(q - new_q) + 4*iabs(r - new_r) + 6*iabs(s - new_s);

            if(gradus>MAX_DISSO)
            {
                event[i]=null;
                 if(LOGGING) console.log("remove (too diss) "+ev[4]);
                return ev;
            }
          }
        }
        return null;
    }

    this.getSize = function()
    {
        var anz=0;
        for(var i=0;i<event.length;i++)
            if(event[i]!=null)
               anz++;
        return anz;
    }

    this.showAll = function()
    {
        var s="all events:\n";
        for(var i=0;i<event.length;i++)
        { 
            var ev = event[i];
            if(ev!=null)
            {
              for(var k=0;k<ev.length;k++)
                 s+=""+ev[k]+" ";
              s+="\n";
            }
        }       
        alert(s);
    }
}





