/**
 * Copyright (c) 2021 Guido Kramann
 *
 * This file is part of a project called kaleidophone - http://www.kramann.info/kaleidophone
 *
 * -------------------------------------------------------
 * This file belongs to Instrument.js (physical modeling).
 * -------------------------------------------------------
 *
 * kaleidophone is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * kaleidophone is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the kaleidophone. If not, see <http://www.gnu.org/licenses/>.
*/

function Oberton()
{
      //Modellparameter, muss erläutert werden
      //  s^2 + Ds +C = 0
      //  =>
      //  s1/2 == -0.5D +/- i*sqrt( C-0.25*D^2 )
      //  Vorgabe: frequenz f, Verhältnis Q = RE/IM, z.B. 1/1000
      //  =>
                //  D = 4*PI*f * Q
                //  C = 0.25*D^2 + (2*PI*f)^2

    var dt=0.0;
    var freq=0.0;
    var amp = 0.0;

    var x=0.0;
    var v=0.0;
    var C=0.0;
    var D=0.0;
  
    function updateCD(Q)
    {
        D = 4.0*Math.PI*freq*Q; 
        C = 0.25*D*D + (2.0*Math.PI*freq)*(2.0*Math.PI*freq);
    }
    this.updateCD = function(Q){updateCD(Q);}    

    this.init = function(Dt,Freq,Q,Amp)
    {
       dt = Dt;
       freq = Freq;
       amp = Amp;
       x = 0.0;
       y = 0.0;
       updateCD(Q);
       //alert("C="+C+" D="+D);
    }

    this.calculateEnergy = function()
    {
        return 0.5*C*x*x + 0.5*D*v*v;
    }

    this.setVtoZero = function()
    {
          v = 0.0;
    }

    this.step = function(F)
    {
       v += (-C*x-D*v+F)*dt;
       x += v*dt; 
       return x;
    }

    this.reset = function()
    {
       x = 0.0;
       v = 0.0;
       F = 0.0;
    }
}

