/**
 * Copyright (c) 2021 Guido Kramann
 *
 * This file is part of a project called kaleidophone - http://www.kramann.info/kaleidophone
 *
 * ---------------------------------------------------------------------------------------------------------
 * Bild is a german word and it means image
 *
 * This file provides methods to draw circles and to project a 3D space with spheres in a plane.
 * 
 * Coordinates of circle for instance are mapped from -1..+1 in both dimensions to the image size in pixels.
 * ---------------------------------------------------------------------------------------------------------
 *
 * kaleidophone is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * kaleidophone is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the kaleidophone. If not, see <http://www.gnu.org/licenses/>.
*/

function Pianoroll()
{
    var TBEGINN = 0; 

    var qapplet;
    var xoff,yoff,width,height;
    var syst;
    var sindex;
    var dtrenner;
    var anz_syst,anz_ticksprosyst;
    var dy,dx,dsy;

    var event=null;
    var event_akt=null;

    //              f f# g g# ...
    var streifen = [0,1,0,1,0,1,0,0,1,0,1,0,   0,1,0,1,0,1,0,0,1,0,1,0,   0,1,0,1,0,1,0,0,1,0,1,0,  0];

    var segmentierung;
    var instrument;

    this.getTPS = function()
    {
        return anz_ticksprosyst;
    }

    this.init = function(xqapplet,xxoff,xyoff,xwidth,xheight, xanz_syst, xanz_ticksprosyst, xdtrenner, umfang,
                         xsegmentierung, xinstrument)
    {
        segmentierung = xsegmentierung;
        instrument = xinstrument;

        qapplet = xqapplet;
        xoff = xxoff;
        yoff = xyoff;
        width = xwidth;
        height = xheight;
        dtrenner = xdtrenner;
        anz_syst = xanz_syst;
        anz_ticksprosyst = xanz_ticksprosyst;
        dx = width/anz_ticksprosyst;
        dsy = height/anz_syst;
//        dy = dsy/(umfang+1);  // mit Lücke
        dy = dsy/(umfang+2);  // mit Lücke

        event_akt = new Array(anz_syst*anz_ticksprosyst);
        for(var i=0;i<event_akt.length;i++)
        {
            event_akt[i] = null;
        }
    }

    this.updateEvents = function(tick_start,B,FACTOR)
    {
        TBEGINN = tick_start;
        var tanz = anz_syst*anz_ticksprosyst;
        event = new Array(tanz);
        for(var i=0;i<tanz;i++)
        {
            var t = i + tick_start;
            var m0 = segmentierung[0].evaluateFormel(true,t+t_ZERO,BASENUMBER,FACTOR,instrument[0].getMidimin(),instrument[0].getMidimax(),instrument[0]);
            var m1 = segmentierung[1].evaluateFormel(true,t+t_ZERO,BASENUMBER,FACTOR,instrument[1].getMidimin(),instrument[1].getMidimax(),instrument[1]);
            //var m2 = segmentierung[2].evaluateFormel(true,t+t_ZERO,BASENUMBER,FACTOR,instrument[2].getMidimin(),instrument[2].getMidimax(),instrument[2]);
            //event[i] = new Array(m0,m1,m2);
            event[i] = new Array(m0,m1);
        }

        for(var i=0;i<tanz;i++)
            event_akt[i] = event[i];
    }

    this.draw = function(TAKT)
    {
        qapplet.noStroke();
        qapplet.fill(0);
        qapplet.rect(xoff,yoff,width,height);
        //qapplet.stroke(255);
        qapplet.fill(255);
        for(var i=0;i<anz_syst;i++)
             qapplet.rect(xoff,yoff+i*dsy+dy,width,dsy-2*dy);
        //qapplet.stroke(200);
        qapplet.fill(200);
        for(var i=0;i<anz_syst;i++)
        {
            var y = yoff + dsy*i;
            for(var ii=0;ii<streifen.length;ii++)
            {
                var yy = y + ii*dy;
                var qq = (ii + streifen.length - midishift)%streifen.length;
                //if(streifen[streifen.length-1-ii]==1)
                if(streifen[streifen.length-1-qq]==1)
                //if(qq==1)
                {
                    qapplet.rect(xoff,yy+dy,width,dy);
                }
            }
        }
    
        //Spielposition:
        var TS = TAKT - TBEGINN;
        if(TS>=0)
        {
            var SYSTEM = Math.floor(TS/anz_ticksprosyst);
            var STELLE = TS%anz_ticksprosyst;
            qapplet.noStroke();
            qapplet.fill(200,200,0,100);
            qapplet.rect(xoff+STELLE*dx,yoff+SYSTEM*dsy+dy,dx,dsy-dy); 
        }

        var ii=0;
        //qapplet.noStroke();
        qapplet.noFill();
        for(var i=0;i<anz_syst;i++)
        {
            var y = yoff + dsy*i;
            for(var k=0;k<anz_ticksprosyst;k++)
            {
                if(event_akt[ii] && event_akt[ii]!=null)
                {
                    for(var kk=0;kk<event_akt[ii].length;kk++)
                    {
                        var mm = event_akt[ii][kk];
                        if(mm && mm!=null && mm.length>0)
                        {
                            //qapplet.fillIndex(kk+1);
                            qapplet.strokeIndex(kk+1);
                            for(var u=0;u<mm.length;u++)
                            {
                                var yyy = y + dsy - dy - dy - dy;
                                var mid = mm[u] - instrument[kk].getMidimin();
                                yyy -= dy*mid;
                                qapplet.rect(xoff+dx*k,yyy+2*kk+dy,dx,dy-4*kk);
                            }
                        }
                    }
                }
                ii++;
            }
        }   
    }
}


