/**
 * Copyright (c) 2021 Guido Kramann
 *
 * This file is part of a project called kaleidophone - http://www.kramann.info/kaleidophone
 *
 * ----------------------------------------------------------------------
 * This file provides the methods for segmentation of the 2-D projections
 * and to derive the AOG formulas from this.
 * Finally tone events are generated with these formulas.
 * ----------------------------------------------------------------------
 *
 * kaleidophone is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * kaleidophone is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the kaleidophone. If not, see <http://www.gnu.org/licenses/>.
*/


function Segmentierung()
{
     var BASIS_AKTUELL = 0;
     //var t_AKTUELL = 0;  // siehe Parameter.js

     var VELOCITY = 0.5;

     var manager;
     var alleinstrumente;

     var midigesamt = "";
     var gebiet;
     var kopie; //Hilfskopie des aktuellen Bildes
     var kopie2; //Hilfskopie des aktuellen Bildes
     var farbanzahl_mit_schwarz;
     var anzahlMax; // maximale Anzahl an Gebieten, die erkannt wird.
     var anzahlAkt; //aktuelle Anzahl zusammenhängender Gebiete 
     var zusammensetzung; // Index0=Anzah der Pixel des Gebiets, dann Anzahl an Flächen mit Index1 usw.

     var mat;
     var mat2;

     var pixanzahl; //für Formel nach Pixanzahl ordnen!
     var inx_pixanzahl; //für Formel nach Pixanzahl ordnen!

     var inst_nr;

     /**
     Index==0 => schwarz, zählt nicht.
     Alles andere bildet zusammenhängende Gebiete
     */
     this.init = function(breite,hoehe,farbanzahl_mit_s,anzMax, xinst_nr, xmanager, xalleinstrumente)
     {
          inst_nr = xinst_nr;

          manager = xmanager;
          alleinstrumente = xalleinstrumente;

          mat = makeArrayZerosZeros(hoehe,breite);
          mat2 = makeArrayZerosZeros(hoehe,breite);
          gebiet = makeArrayZerosZeros(hoehe,breite);
          kopie = makeArrayZerosZeros(hoehe,breite);
          kopie2 = makeArrayZerosZeros(hoehe,breite);
          farbanzahl_mit_schwarz = farbanzahl_mit_s;
          anzahlMax = anzMax;
          zusammensetzung = makeArrayZerosZeros(anzMax,farbanzahl_mit_schwarz);
          pixanzahl = makeZeros(anzahlMax);
          inx_pixanzahl = makeZeros(anzahlMax);
          anzahlAkt=0;
          //alert("mat.length"+mat.length);
     }
     
     this.updateZusammensetzung = function(m)
     {
          var ii=0;
//alert("mat.length="+mat.length+" mat[0].length="+mat[0].length);
          for(var i=0;i<mat.length;i++)
          {
              for(var k=0;k<mat[i].length;k++)
              {
                  mat[i][k] = m[ii];
                  // **** NEU ****
                  if(mat[i][k]>5) //ggf. Schwarz-Index 6 zu 0 setzen
                  {
                      mat[i][k] = 0; 
                  }
                  ii = ii + 1;
              }
          }

          //Blur hart coden: NEIN, Hilft nichts!!!
          /*
          for(var i=0;i<mat.length;i++)
          {
              for(var k=0;k<mat[i].length;k++)
              {
                  mat[i][k] = 0;
              }          
          }
          for(var CC=1;CC<=5;CC++)
          {
              for(var i=0;i<mat.length;i++)
              {
                  for(var k=0;k<mat[i].length;k++)
                  {
                      if(mat2[i][k] == CC)
                      {
                          mat[i][k] = CC;
                          if(i<mat.length-1) mat[i+1][k]=CC;
                          if(i>0) mat[i-1][k]=CC;
                          if(k<mat[0].length-1) mat[i][k+1]=CC;
                          if(k>0) mat[i][k-1]=CC;

                          if(i<mat.length-1 && k<mat[0].length-1) mat[i+1][k+1]=CC;
                          if(i>0 && k<mat[0].length-1) mat[i-1][k+1]=CC;
                          if(k>0 && i<mat.length-1) mat[i+1][k-1]=CC;
                          if(k>0 && i>0) mat[i-1][k-1]=CC;


                      }
                  }          
              }
          }
          */
          updateZusammensetzungMat();        
     }
     //this.updateZusammensetzung = function(m) {updateZusammensetzung(m);}    

     function updateZusammensetzungMat()
     {
          if(mat.length!=gebiet.length || mat[0].length!=gebiet[0].length)
          {
              gebiet = makeArrayZerosZeros(mat.length,mat[0].length);
              kopie = makeArrayZerosZeros(mat.length,mat[0].length);
              kopie2 = makeArrayZerosZeros(mat.length,mat[0].length);
          }    
          for(var i=0;i<gebiet.length;i++)
              for(var k=0;k<gebiet[i].length;k++)
                  gebiet[i][k]=0;
          for(var i=0;i<kopie.length;i++)
              for(var k=0;k<kopie[i].length;k++)
                  kopie[i][k]=mat[i][k];
                  
          var zaehlung = 1;
          anzahlAkt=0;
          for(var i=0;i<kopie.length;i++)
          {
               for(var k=0;k<kopie[i].length;k++)
               {
                     if(kopie[i][k]>0 && anzahlAkt<anzahlMax) //Neues Gebiet gefunden
                     {
                          //gebiet[i][k] = zaehlung;
                          
                          //Alle mit dem Pixel zusammenhängende anderen Pixel aus kopie entfernen und
                          //in gebiet mit zaehlung versehen:
                          
                          setzeUmgebung(i,k,kopie,gebiet,zaehlung);
                          
                          for(var ii=0;ii<kopie2.length;ii++)
                              for(var kk=0;kk<kopie2[i].length;kk++)
                                   kopie2[ii][kk]=0;
                          for(var ii=0;ii<kopie2.length;ii++)
                              for(var kk=0;kk<kopie2[i].length;kk++)
                                   if(gebiet[ii][kk]==zaehlung)//nur aktuell betrachtetes gebiet hernehmen
                                        kopie2[ii][kk]=mat[ii][kk];
                          var groesse = 0;
                          for(var ii=0;ii<kopie2.length;ii++)
                              for(var kk=0;kk<kopie2[ii].length;kk++)
                                  if(kopie2[ii][kk]>0)
                                     groesse++;
                          zusammensetzung[anzahlAkt][0]=groesse;           
                          zaehleFelder(zusammensetzung[anzahlAkt],kopie2);
                          zaehlung++;
                          anzahlAkt++;
                     }
               }
          }
     }
     this.updateZusammensetzungMat = function(mat){updateZusammensetzungMat(mat);}

     var ENTFERNT=0;
     function zaehleFelder(statistik,kopie2)
     {
              for(var i=1;i<statistik.length;i++)
                 statistik[i]=0;
              //für jeden Pixel identifizieren und deren Umgebung gleicher farbe rekursiv löschen:
              for(var i=0;i<kopie2.length;i++)
              {
                  for(var k=0;k<kopie2[i].length;k++)
                  {
                       if(kopie2[i][k]>0) //Schwelle eingeführt!!!!!
                       {
                           var finx = kopie2[i][k]; //Farbindex merken
                           statistik[finx]++; // wieder ein Bereich/Fleck mehr in dieser Farbe
                           ENTFERNT=0;
                           entferneFarbe(i,k,finx,kopie2);
                           if(ENTFERNT<MINDEST_PIXEL_PRO_FELD)
                           {
                               //Rückgängig machen, übergehen!!!!
                               statistik[finx]--;
                           }
                       }
                  }
              }
     }

     function  entferneFarbe(i,k,finx,kopie2)
     {
         kopie2[i][k]=0;
         ENTFERNT++;
         if(i>0 && kopie2[i-1][k]==finx) entferneFarbe(i-1,k,finx,kopie2);
         if(k>0 && kopie2[i][k-1]==finx) entferneFarbe(i,k-1,finx,kopie2);
         if(i<kopie2.length-1 && kopie2[i+1][k]==finx) entferneFarbe(i+1,k,finx,kopie2);
         if(k<kopie2[0].length-1 && kopie2[i][k+1]==finx) entferneFarbe(i,k+1,finx,kopie2);
     }

     
     function setzeUmgebung(i,k,kopie,gebiet,zaehlung)
     {
          gebiet[i][k] = zaehlung;
          kopie[i][k] = 0;
          
          //nur orthogonal angrenzende, oder auch diagonal? -- ausprobieren
          if(i>0 && kopie[i-1][k]>0) setzeUmgebung(i-1,k,kopie,gebiet,zaehlung);
          if(i<kopie.length-1 && kopie[i+1][k]>0) setzeUmgebung(i+1,k,kopie,gebiet,zaehlung);
          if(k>0 && kopie[i][k-1]>0) setzeUmgebung(i,k-1,kopie,gebiet,zaehlung);
          if(k<kopie[0].length-1 && kopie[i][k+1]>0) setzeUmgebung(i,k+1,kopie,gebiet,zaehlung);
     }
     
     this.drawGebiete = function(qapplet,xoff,yoff,width,height)
     {
          var dx = width/gebiet[0].length;
          var dy = height/gebiet.length;
          
          for(var i=0;i<gebiet.length;i++)
          {
              for(var k=0;k<gebiet[i].length;k++)
              {
                   var c = gebiet[i][k];
                   if(c==0)
                   {
                       qapplet.fill(0);
                       qapplet.stroke(0);
                   }
                   else
                   {
                       var R = 127+127*((c/1)%2);
                       var G = 127+127*((c/2)%2);
                       var B = 127+127*((c/4)%2);
                       qapplet.fill(R,G,B);
                       qapplet.stroke(R,G,B);
                   }
                   qapplet.rect(xoff+k*dx,yoff+i*dy,dx,dy);
              }
          }
     }
     
     this.zeigeGebietsgroessen = function()
     {
           var s="";
           for(var i=0;i<anzahlAkt;i++)
              s+=(zusammensetzung[i][0]+" ");
           console.log(s);   
     }
     
     this.zeigeGebietsstruktur = function()
     {
           console.log("Struktur:\n");
           for(var i=0;i<anzahlAkt;i++)
           {
              var s="";
              for(var k=1;k<zusammensetzung[i].length;k++)
                 s+=(zusammensetzung[i][k]+" ");
              s+=(zusammensetzung[i][0]+" ");
              console.log(s);   
           }   
           console.log("\n");   
     }

     //************************************
     //Ergänzungen, um Formeln zu gewinnen:
     //************************************

     
     function getFormel()
     {
         var s="";
         if(anzahlAkt>0)
         {
            for(var i=0;i<anzahlAkt;i++)
               pixanzahl[i] = zusammensetzung[i][0];
            sortIndicesBiggestFirst(pixanzahl,anzahlAkt,inx_pixanzahl);
            for(var ii=0;ii<anzahlAkt;ii++)
            {
                 var i = inx_pixanzahl[ii];
                 //wenn graue dabei:
                 if(zusammensetzung[i][5]>0)
                 {
                     var opindex = (zusammensetzung[i][5]-1)%5; // * / % + -
                     var op="";
                     if(opindex==0) op="%";
                     if(opindex==1) op="/";
                     if(opindex==2) op="*";
                     if(opindex==3) op="+";
                     if(opindex==4) op="-";
                     var x=0;
                     for(var k=0;k<zusammensetzung[i][1];k++) x+=2;
                     for(var k=0;k<zusammensetzung[i][2];k++) x+=3;
                     for(var k=0;k<zusammensetzung[i][3];k++) x+=5;
                     for(var k=0;k<zusammensetzung[i][4];k++) x+=7;
                     if(x>1)
                     {
                          s+=op+x;
                     }
                 }
            }
         }   
         return s;
     }     
     this.getFormel = function(){return getFormel();}

     this.printFormel = function(i)
     {
          var s = getFormel();
          if(s.length>1)
              console.log("Formel Nr."+i+": "+s);
     }

     this.showFormel = function(qapplet,xoff,yoff)
     {
          var s = ""+BASIS_AKTUELL+"//("+t_AKTUELL+getFormel()+")";
          if(s.length>1)
          {
              //qapplet.fill(0);
              if(midigesamt!="")
              {
                  //qapplet.text(s+"->"+midigesamt,xoff,yoff);
                  qapplet.text(s,xoff,yoff);
                  qapplet.text("->"+midigesamt,xoff,yoff+14);
              }
              else
              {
                  qapplet.text(s,xoff,yoff);
              }
          }    
     }

     function xRoundrobin(x,B)
     {
         var p = 0;
         while(B>=2 && B%2==0) {B/=2; p++;}
         var q = 0;
         while(B>=3 && B%3==0) {B/=3; q++;}
         var r = 0;
         while(B>=5 && B%5==0) {B/=5; r++;}
         var v = 0;
         while(B>=7 && B%7==0) {B/=7; v++;}

         var pp = 0;
         while(x>=2 && x%2==0) {x/=2; pp++;}
         var qq = 0;
         while(x>=3 && x%3==0) {x/=3; qq++;}
         var rr = 0;
         while(x>=5 && x%5==0) {x/=5; rr++;}
         var vv = 0;
         while(x>=7 && x%7==0) {x/=7; vv++;}

         while(pp>p) pp-=p;
         while(qq>q) qq-=q;
         while(rr>r) rr-=r;
         while(vv>v) vv-=v;

         var xneu = 1;
         for(var i=0;i<pp;i++) xneu*=2;
         for(var i=0;i<qq;i++) xneu*=3;
         for(var i=0;i<rr;i++) xneu*=5;
         for(var i=0;i<vv;i++) xneu*=7;

         return xneu;
     }
     this.xRoundrobin = function(x,B){return xRoundrobin(x,B);}

     this.evaluateFormel = function(SILENT,t,B,FACTOR,midimin,midimax,INSTRUMENT)  //Nach und nach weitere Parameter, um varianten zu erzeugen!!!!!!!
     {
         var SMIDI;
         var SINX=0;

         if(anzahlAkt>0 && MULTIPLY_BASE_WITH_AREA_COUNT==true)
         {
             B = B * anzahlAkt;
         }
         BASIS_AKTUELL = B;  // für Anzeigen der Formel
         //t_AKTUELL = t;  // für Anzeigen der Formel, siehe Player.js

         var pot = number2potency(B);

         midigesamt = "";

         if(SILENT) SMIDI = new Array();
      

         if(anzahlAkt>0)
         {
            for(var i=0;i<anzahlAkt;i++)
               pixanzahl[i] = zusammensetzung[i][0];
            sortIndicesBiggestFirst(pixanzahl,anzahlAkt,inx_pixanzahl);
            
            var xx = t;
            
            for(var ii=0;ii<anzahlAkt;ii++)
            {
                 var i = inx_pixanzahl[ii];
                 //wenn graue dabei:
                 if(zusammensetzung[i][5]>0)
                 {
                     var x=0;
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                     for(var k=0;k<zusammensetzung[i][1];k++) x+=2;  //zukünftig auf Produkt/Multiplikation der zaheln umschaltbar machen !!!!!!
                     for(var k=0;k<zusammensetzung[i][2];k++) x+=3;
                     for(var k=0;k<zusammensetzung[i][3];k++) x+=5;
                     for(var k=0;k<zusammensetzung[i][4];k++) x+=7;
                     if(x>1)
                     {
//!!!!!!!!!!
                         var opindex = (zusammensetzung[i][5]-1)%5; // * / % + -    RANGFOLGE steuerbar machen!!!!!!!!!!!
                         var op="";
                         if(opindex==0) xx%=x;
                         if(opindex==1) xx/=x;
                         if(opindex==2) xx*=x;
                         if(opindex==3) xx+=x;
                         if(opindex==4) xx-=x;

                         //HIER SPIELEN
                         //HIER SPIELEN
                         //HIER SPIELEN
            if(xx!=t) //wenn verändert
            {
                //*****************************
                var xxneu = xRoundrobin(xx,B);
                //*****************************

                //xx = xx%B;
                //if(xx<1) xx=1;

                var freq = B;
                var f = 1;
                var p=0;
                var q=0;
                var r=0;
                var v=0;
                var pp=pot[0];
                var qq=pot[1];
                var rr=pot[2];
                var vv=pot[3];
                while(freq>=2 && freq%2==0 && xxneu>=2 && xxneu%2==0) {freq/=2;xxneu/=2;f*=2;p++;pp--;}
                while(freq>=3 && freq%3==0 && xxneu>=3 && xxneu%3==0) {freq/=3;xxneu/=3;f*=3;q++;qq--;}
                while(freq>=5 && freq%5==0 && xxneu>=5 && xxneu%5==0) {freq/=5;xxneu/=5;f*=5;r++;rr--;}
                while(freq>=7 && freq%7==0 && xxneu>=7 && xxneu%7==0) {freq/=7;xxneu/=7;f*=7;v++;vv--;}
                
                var midi = Math.round(frequency2Midi(FACTOR*freq));
                var midi2 = Math.round(frequency2Midi(FACTOR*f));
                
                if(midi>=midimin && midi<=midimax) // freq geteilt
                {
                     //return midi;
                  if(SILENT)
                  {
                      SMIDI[SINX++] = midi;
                  }
                  else
                  {
                     var ANZ=0;
                     do
                     {
                         //this.returnAndRemoveTooDissonant = function(MAX_DISSO,   new_p,new_q,new_r,new_s)
                         var ev = manager.returnAndRemoveTooDissonant(MAX_DISSONANT,pp,qq,rr,vv);
                         if(ev!=null)
                         {
                             var INR = ev[3];
                             alleinstrumente[INR].damp(ev[4]); 
                             ANZ++;              
                         }
                     } while(ev!=null);
                     manager.addEvent(t,inst_nr,midi,pp,qq,rr,vv);

                     midigesamt+=" "+midi;

                     if(ANZ>0)
                         VELOCITY += ANZ*0.1;
                     else
                         VELOCITY -= 0.1;
                     while(VELOCITY>1.2) VELOCITY-=0.6;
                     while(VELOCITY<0.2) VELOCITY+=0.6;

                     if(MUTE==false) INSTRUMENT.undamp(midi);
                     if(MUTE==false) INSTRUMENT.play(midi,VELOCITY);
//instrumentWav[0].play(midi,1.0);
                  }
                }
                else if(midi2>=midimin && midi2<=midimax) // f mal
                {
                  if(SILENT)
                  {
                      SMIDI[SINX++] = midi2;
                  }
                  else
                  {
                     //return midi2;
                     do
                     {
                         //var ev = manager.returnAndRemoveOldest();
                         var ev = manager.returnAndRemoveTooDissonant(MAX_DISSONANT,p,q,r,v);
                         if(ev!=null)
                         {
                             var INR = ev[3];
                             alleinstrumente[INR].damp(ev[4]);               
                         }
                     } while(ev!=null);
                     manager.addEvent(t,inst_nr,midi2,p,q,r,v);
                     midigesamt+=" "+midi2;


                     if(ANZ>0)
                         VELOCITY += ANZ*0.2;
                     else
                         VELOCITY -= 0.2;
                     while(VELOCITY>1.6) VELOCITY-=0.9;
                     while(VELOCITY<0.3) VELOCITY+=0.9;

                     if(MUTE==false) INSTRUMENT.undamp(midi2);
                     if(MUTE==false) INSTRUMENT.play(midi2,VELOCITY);
//instrumentWav[0].play(midi2,1.0);
                  }
                }
            }

                         //ENDE HIER SPIELEN
                         //ENDE HIER SPIELEN
                         //ENDE HIER SPIELEN
                     }
                 }
            }//for

            if(SILENT)
                return SMIDI;
            else
            {
                var V = Math.round(VELOCITY*100.0);
                if(V<20) V=20;
                if(V>99) V=99;
                if(MUTE==false)
                   return V+" | "+midigesamt;
                else
                   return "";
            }
          /*
            if(xx!=t) //wenn verändert
            {
                //*****************************
                var xxneu = xRoundrobin(xx,B);
                //*****************************

                //xx = xx%B;
                //if(xx<1) xx=1;

                var freq = B;
                var f = 1;
                while(freq>=2 && freq%2==0 && xxneu>=2 && xxneu%2==0) {freq/=2;xxneu/=2;f*=2;}
                while(freq>=3 && freq%3==0 && xxneu>=3 && xxneu%3==0) {freq/=3;xxneu/=3;f*=3;}
                while(freq>=5 && freq%5==0 && xxneu>=5 && xxneu%5==0) {freq/=5;xxneu/=5;f*=5;}
                while(freq>=7 && freq%7==0 && xxneu>=7 && xxneu%7==0) {freq/=7;xxneu/=7;f*=7;}
                
                var midi = Math.round(frequency2Midi(FACTOR*freq));
                var midi2 = Math.round(frequency2Midi(FACTOR*f));
                
                if(midi>=midimin && midi<=midimax)
                {
                     return midi;
                }
                else if(midi2>=midimin && midi2<=midimax)
                {
                     return midi2;
                }
                else
                {
                     return 0;
                }
            }
            else
            {
                return 0;
            }
           */
         }
         else // if(anzahlAkt>0)
         {
            if(SILENT)
                return SMIDI;
            else
                return "";
         }
       
     }
     
     
     this.evaluateFormelXX = function(t)
     {
         if(anzahlAkt>0)
         {
            for(var i=0;i<anzahlAkt;i++)
               pixanzahl[i] = zusammensetzung[i][0];
            sortIndicesBiggestFirst(pixanzahl,anzahlAkt,inx_pixanzahl);
            
            var xx = t;
            
            for(var ii=0;ii<anzahlAkt;ii++)
            {
                 var i = inx_pixanzahl[ii];
                 //wenn graue dabei:
                 if(zusammensetzung[i][5]>0)
                 {
                     var x=0;
                     for(var k=0;k<zusammensetzung[i][1];k++) x+=2;
                     for(var k=0;k<zusammensetzung[i][2];k++) x+=3;
                     for(var k=0;k<zusammensetzung[i][3];k++) x+=5;
                     for(var k=0;k<zusammensetzung[i][4];k++) x+=7;
                     if(x>1)
                     {
                         var opindex = (zusammensetzung[i][5]-1)%5; // * / % + -
                         var op="";
                         if(opindex==0) xx%=x;
                         if(opindex==1) xx/=x;
                         if(opindex==2) xx*=x;
                         if(opindex==3) xx+=x;
                         if(opindex==4) xx-=x;
                     }
                 }
            }//for
            return xx;            
         }
         else // if(anzahlAkt>0)
         {
            return t;
         }
     }         
}
