/**
 * Copyright (c) 2021 Guido Kramann
 *
 * This file is part of a project called kaleidophone - http://www.kramann.info/kaleidophone
 *
 * --------------------------------------------------------------------
 * Szene allows to get an image as a projection of a 3-D scene.
 * The 3-D scene consists of colored spheres all of the same size only.
 * --------------------------------------------------------------------
 *
 * kaleidophone is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * kaleidophone is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the kaleidophone. If not, see <http://www.gnu.org/licenses/>.
*/

function Szene()
{
    var COLOR = [
        [0,0,0],     // 0 black
        [0,0,255],   // 1 blue    represents the prime number 2
        [255,0,0],   // 2 red     represents the prime number 3   
        [0,255,0],   // 3 green   represents the prime number 5
        [255,255,0], // 4 yellow  represents the prime number 7
        [127,127,127],// 5 grey    multiple grey fields represent an operator
        [31,31,31]];    // 6 black to hide things, bei Segmentierung: alle Indices AUSSER 6 als Farbe zusammenfassen!
//Der Einfachheit den Index 6 wenn er auftaucht VOR der Segmentierung in Index 0 verwandeln

 //BILD:
    //int[] m; //represents the image
    var m;
    //int[] n; //same size as m. Used for buffering and pre-calculation.
    var n;
    var width,height;
    

    this.getM = function()
    {
        return m;
    }

    this.getWidth = function()
    {
        return width;
    }

    this.getHeight = function()
    {
        return height;
    }

    function clearM()
    {
             for(var i=0;i<m.length;i++) m[i]=0;
    }
    this.clearM  = function() {clearM();}

    function clear(q)
    {
             for(var i=0;i<q.length;i++) q[i]=0;
    }
    this.clear  = function(q) {clear(q);}

    
    function xpix2x(x){ return  -1.0 + 2.0 * x / width;}
    function ypix2y(y){ return   1.0 - 2.0 * y / height;}
    function x2xpix(x){ return  Math.floor(0.5*(x+1.0)*width);}
    function y2ypix(y){ return  Math.floor(0.5*(-y+1.0)*height);}
    function inx2x(inx){ return  inx%width;}
    function inx2y(inx){ return  Math.floor(inx/width);}
    
    function circle(x,y,radius,colcode)
    {
        for(var i=0;i<m.length;i++)
        {
            var x2 = xpix2x(inx2x(i));
            var y2 = ypix2y(inx2y(i));
            var dist = Math.sqrt((x-x2)*(x-x2)+(y-y2)*(y-y2));
            if(dist<=radius) m[i]=colcode;
        }
    }
    this.circle = function(x,y,radius,colcode){circle(x,y,radius,colcode);};
    /*
    this.draw = function(qapplet,xoff,yoff,breite,hoehe)
    {
         var dx = breite/width;
         var dy = hoehe/height;
         qapplet.noStroke();
         for(var i=0;i<m.length;i++)
         {
              var c = m[i];
              var x = inx2x(i);
              var y = inx2y(i);
              //qapplet.fill(COLOR[c][0],COLOR[c][1],COLOR[c][2]);
              qapplet.fillIndex(c);
              qapplet.rect(xoff+x*dx,yoff+y*dy,dx,dy);
         }
    }
    */
    this.draw = function(qapplet,xoff,yoff,breite,hoehe)
    {
         //var dx = breite/width;
         //var dy = hoehe/height;
         qapplet.noStroke();
         for(var i=0;i<m.length;i++)
         {
              var c = m[i];
              var x = inx2x(i);
              var y = inx2y(i);
              //qapplet.fill(COLOR[c][0],COLOR[c][1],COLOR[c][2]);
              qapplet.fillIndex(c);
              qapplet.rect(xoff+x,yoff+y,1,1);
         }
    }

    this.drawPixel = function(qapplet,xoff,yoff,breite,hoehe,pix)
    {
         //var dx = breite/width;
         //var dy = hoehe/height;
         //qapplet.noStroke();
         for(var i=0;i<m.length;i++)
         {
              var c = m[i];

              //var x = i%KANTE;
              //var y = Math.floor(i/KANTE);

              pix[i*4+0] = COLOR[c][0];
              pix[i*4+1] = COLOR[c][1];
              pix[i*4+2] = COLOR[c][2];
              pix[i*4+3] = 255;

            /*              
              var index1 = (y*dPIXEL+0)*dKANTE + (x*dPIXEL+0);
              var index2 = (y*dPIXEL+0)*dKANTE + (x*dPIXEL+1);
              var index3 = (y*dPIXEL+1)*dKANTE + (x*dPIXEL+0);
              var index4 = (y*dPIXEL+1)*dKANTE + (x*dPIXEL+1);

              pix[index1*4+0] = COLOR[c][0];
              pix[index1*4+1] = COLOR[c][1];
              pix[index1*4+2] = COLOR[c][2];
              pix[index1*4+3] = 255;

              pix[index2*4+0] = COLOR[c][0];
              pix[index2*4+1] = COLOR[c][1];
              pix[index2*4+2] = COLOR[c][2];
              pix[index2*4+3] = 255;

              pix[index3*4+0] = COLOR[c][0];
              pix[index3*4+1] = COLOR[c][1];
              pix[index3*4+2] = COLOR[c][2];
              pix[index3*4+3] = 255;

              pix[index4*4+0] = COLOR[c][0];
              pix[index4*4+1] = COLOR[c][1];
              pix[index4*4+2] = COLOR[c][2];
              pix[index4*4+3] = 255;
            */
         }
    }

  //SZENE

    //int[][][] s;
    var s;
    //double[][][][] koord;
    var koord;

    //double[][][][] tkoord; //Transformierte Koordinaten
    var tkoord;

    var radius; //0..1, Ausdehnung in eine Achsrichtung des Kubus', der alle Kugeln enthält
    var r; //Radius einer einzelnen Kugel
    
    //Reihenfolge, in der gezeichnet werden muss:
    //int[] iindexT;// = new int[wT.length];
    var iindexT;
    //double[] zzT;// = new int[wT.length];
    var zzT;    

    this.getS = function()
    {
        return s;
    }

    this.init = function(width_and_height, rt, rs)
    {
        //super(width_and_height);
        width = width_and_height;
        height = width_and_height;
        m = makeZeros(width*height);
        n = makeZeros(width*height);

           
        s = makeArrayZerosZerosZeros(rt,rt,rt);
        radius = rs;
        r = 0.5*(2.0*radius)/rt;
        //Mittelpunktskoordinate jeder Stelle
        koord = makeArrayZerosZerosZerosZeros(rt,rt,rt,3);
        tkoord = makeArrayZerosZerosZerosZeros(rt,rt,rt,3);
        zzT = makeZeros(rt*rt*rt);
        iindexT = makeZeros(rt*rt*rt);
        bestimmeKoordinaten();
    }
    
    function bestimmeKoordinaten()
    {
        var delta = (radius*2.0) / koord.length;
        var offset = -radius + 0.5*delta;
        for(var z=0;z<koord.length;z++)
        {
             var zz = offset + z*delta;
             for(var y=0;y<koord[z].length;y++)
             {
                   var yy = offset + y*delta;
                   for(var x=0;x<koord[z][y].length;x++)
                   {
                       var xx = offset + x*delta;
                       koord[z][y][x][0] = xx; 
                       koord[z][y][x][1] = yy; 
                       koord[z][y][x][2] = zz; 
                   }
             }
        }
    }
    
    function projiziereSzene(rotX,rotY)
    {
        var ii=0;
        for(var z=0;z<koord.length;z++)
        {
             for(var y=0;y<koord[z].length;y++)
             {
                 for(var x=0;x<koord[z][y].length;x++)
                 {
                     var xx = koord[z][y][x][0];
                     var yy = koord[z][y][x][1];
                     var zz = koord[z][y][x][2];
                               
                     var xT = xx*Math.cos(rotY)-zz*Math.sin(rotY);
                     var yT = yy;
                     var zT = xx*Math.sin(rotY)+zz*Math.cos(rotY);

                     tkoord[z][y][x][0] = xT;
                     tkoord[z][y][x][1] = yT*Math.cos(rotX) - zT*Math.sin(rotX);
                     tkoord[z][y][x][2] = yT*Math.sin(rotX) + zT*Math.cos(rotX);
        
                     zzT[ii] = tkoord[z][y][x][2];
                     ii++;
                 }
             }
        }
        sortIndicesBiggestFirst(zzT,zzT.length,iindexT);
        clearM();
        for(var i=0;i<iindexT.length;i++)
        {
             var inx = iindexT[i];
             var x=inx%koord.length;
             var y=(Math.floor(inx/koord.length))%koord.length;
             var z=Math.floor(inx/(koord.length*koord.length));
             
             if(s[z][y][x]>0)
             {
                var rr = r - 0.025*tkoord[z][y][x][2]; //vorn groesser
                circle(tkoord[z][y][x][0],tkoord[z][y][x][1],rr,s[z][y][x]);
             }
        }
        
        
        
    }
    this.projiziereSzene = function(rotX,rotY){projiziereSzene(rotX,rotY);}
    
    function set(x, y, z, colcode)
    {
        s[z][y][x]=colcode;
    }
    this.set = function(x, y, z, colcode){set(x, y, z, colcode);}

    this.getS = function()
    {
       return s;
    }

    this.copyStoHere = function(sneu)
    {
        for(var i=0;i<s.length;i++)
            for(var k=0;k<s[i].length;k++)
                for(var q=0;q<s[i][k].length;q++)
                    s[i][k][q] = sneu[i][k][q];
    }

    this.drawKaleidoskopisch = function(qapplet,xoff,yoff,breite,hoehe,ex,ey,eshift,ALFA)
    {
         var dx = breite/width;
         var dy = hoehe/height;
         qapplet.noStroke();
         for(var i=0;i<m.length;i++)
         {
              var c = m[i];
              var x = inx2x(i);
              var y = inx2y(i);
              
              var xx = x*dx*ex[0] + y*dy*ey[0];
              var yy = x*dx*ex[1] + y*dy*ey[1];
              
              xx += eshift[0]*KANTE;
              yy += eshift[1]*KANTE;
              
              if(c!=0)
              {
                  qapplet.fill(COLOR[c][0],COLOR[c][1],COLOR[c][2],ALFA);
              //rect(xoff+(float)x*dx,yoff+(float)y*dy,dx,dy);
                  qapplet.rect(xoff+xx,yoff+yy,dx,dy);
              }
         }
    }

    this.drawKaleidoskopischContext = function(context,xoff,yoff,breite,hoehe,ex,ey,eshift,ALFA,CINX,WIDTH,HEIGHT)
    {
         var dx = 1.5*(breite/width);
         var dy = 1.5*(hoehe/height);

         //qapplet.noStroke();
         for(var i=0;i<m.length;i++)
         {
              var c = m[i];
              var x = inx2x(i);
              var y = inx2y(i);
              
              var xx = x*dx*ex[0] + y*dy*ey[0];
              var yy = x*dx*ex[1] + y*dy*ey[1];
              
              xx += eshift[0]*KANTE;
              yy += eshift[1]*KANTE;
              
              if(c!=0)
              {

                  //  translate(width/2 - 1.5*width/2,height/2 - 1.5*height/2);
                  //  scale(1.5,1.5);  //siehe setup()


                  //qapplet.fill(COLOR[c][0],COLOR[c][1],COLOR[c][2],ALFA);
                  context.fillStyle = CINX[c];
              //rect(xoff+(float)x*dx,yoff+(float)y*dy,dx,dy);
                  //qapplet.rect(xoff+xx,yoff+yy,dx,dy);
                  context.fillRect(xoff+xx,yoff+yy,dx,dy);
              }
         }
    }

    this.drawKaleidoskopischContextNeu = function(qapplet,context,xoff,yoff,breite,hoehe,ex,ey,eshift,ALFA,CINX,WIDTH,HEIGHT)
    {
         var dx = 1.5*(breite/width);
         var dy = 1.5*(hoehe/height);
         qapplet.noStroke();

         for(var i=0;i<iindexT.length;i++)
         {
             var inx = iindexT[i];
             var x=inx%koord.length;
             var y=(Math.floor(inx/koord.length))%koord.length;
             var z=Math.floor(inx/(koord.length*koord.length));
             var c = s[z][y][x];
             var QQ = height*1.5;
             var RR = height/10.0;
             var xxx = tkoord[z][y][x][0];
             var yyy = tkoord[z][y][x][1];

             if(c>0)
             {
                var xx = QQ*xxx*ex[0] + QQ*yyy*ey[0];
                var yy = QQ*xxx*ex[1] + QQ*yyy*ey[1];
                //xx += eshift[0]*KANTE;
                //yy += eshift[1]*KANTE;
                var rr = r - 0.025*tkoord[z][y][x][2]; //vorn groesser
                qapplet.fillIndexA2(c);
                qapplet.circle(xoff+xx,yoff+yy,RR);
             }
         }

    }

    function setKoord(x,y,breite,pix,cinx,ALFA)
    {
        var ii = Math.floor(x + breite*y);
        if(ii>=0 && ii*4+4<pix.length)
        {
          pix[ii*4+0] = COLOR[cinx][0];
          pix[ii*4+1] = COLOR[cinx][1];
          pix[ii*4+2] = COLOR[cinx][2];
          pix[ii*4+3] = ALFA;
        }
    }

    this.drawKaleidoskopischPixel = function(qapplet,xoff,yoff,breite,hoehe,ex,ey,eshift,ALFA,pix)
    {
         var dx = 1.0;//breite/width;
         var dy = 1.0;//hoehe/height;
         qapplet.noStroke();

         for(var i=0;i<m.length;i++)
         {
              var c = m[i];
              var x = inx2x(i);
              var y = inx2y(i);
              
              var xx = x*dx*ex[0] + y*dy*ey[0];
              var yy = x*dx*ex[1] + y*dy*ey[1];
              
              xx += eshift[0]*KANTE;
              yy += eshift[1]*KANTE;
              
              if(c!=0)
              {
                  setKoord(xoff+xx,yoff+yy,breite,pix,c,ALFA);
/*
                  var ii = Math.floor(xoff+xx+0.5*hoehe + breite*(yoff+yy+0.5*hoehe)); 

                  if(ii>=0 && ii*4+4<pix.length)
                  {
                      pix[ii*4+0] = COLOR[c][0];
                      pix[ii*4+1] = COLOR[c][1];
                      pix[ii*4+2] = COLOR[c][2];
                      pix[ii*4+3] = 255;
                  }
*/
                  //qapplet.fill(COLOR[c][0],COLOR[c][1],COLOR[c][2],ALFA);
              //rect(xoff+(float)x*dx,yoff+(float)y*dy,dx,dy);
                  //qapplet.rect(xoff+xx,yoff+yy,dx,dy);
              }
         }

         //for(var x=0;x<10;x++)
         //    for(var y=0;y<20;y++)
         //          setKoord(x,y,breite,pix,1);
    }


}

