/**
 * Copyright (c) 2021 Guido Kramann
 *
 * This file is part of a project called kaleidophone - http://www.kramann.info/kaleidophone
 *
 * ------------------------------------------------------------------------------------------------------
 * To provide sound for less powerful computers Wave generates look up tables from a given physical 
 * modeling instrument.
 * ------------------------------------------------------------------------------------------------------
 *
 * kaleidophone is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * kaleidophone is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the kaleidophone. If not, see <http://www.gnu.org/licenses/>.
*/

function Wave()
{
    var instrument;
    var sample_rate;
    var velocity;
    var minenergy;
    var mmin,mmax;
    this.init = function(xinstrument,xsample_rate,xvelocity,xminenergy)
    {
        instrument = xinstrument;
        sample_rate = xsample_rate;
        velocity = xvelocity;
        minenergy = xminenergy;
        mmin = instrument.getMidimin();
        mmax = instrument.getMidimax();
    }

    /**
        Returns samples in this array format:
        arr[2*(midi-midimin)+0] = left
        arr[2*(midi-midimin)+1] = right  
    */
    this.generateWave = function()
    {
        var arr = new Array(2*(mmax-mmin+1));
        var active = new Array(2*(mmax-mmin+1));
        var xalt = new Array(2*(mmax-mmin+1));

        for(var i=0;i<arr.length;i+=2)
        {
            var midi = Math.floor(i/2) + mmin;
            instrument.undamp(midi);
            instrument.play(midi,velocity);  

            active[i] = true;
            active[i+1] = true;          
            xalt[i] = 0.0;
            xalt[i+1] = 0.0;          
            arr[i] = new Array();
            arr[i+1] = new Array();
        }

        var index=0;
        var gefunden = true;
        while(gefunden)  //alle Töne gleichzeitig spielen, bzw. berechnen
//        for(var ii=0;ii<10000;ii++)  //alle Töne gleichzeitig spielen, bzw. berechnen
        {
            gefunden = false;
            for(var i=0;i<arr.length;i+=2)
            {

                if(active[i]==true || active[i+1]==true)
                {
                    var midi = Math.floor(i/2) + mmin;
                    var energy = instrument.calculateEnergy(midi);
                    if(active[i]==true)
                    {
                        gefunden = true;
                        arr[i][index] = instrument.getLLEFT(midi);


                        if(index>1000 && energy<minenergy && arr[i][index]*xalt[i]<=0.0)
                        {
                            active[i] = false;
                        }

                        xalt[i] = arr[i][index];                      

                    }                    

                    if(active[i+1]==true)
                    {
                        gefunden = true;
                        arr[i+1][index] = instrument.getRRIGHT(midi);

                        if(index>1000 && energy<minenergy && arr[i+1][index]*xalt[i+1]<=0.0)
                        {
                            active[i+1] = false;
                        }

                        xalt[i+1] = arr[i+1][index];                      
                    }                    
                }

            }
            instrument.step();
            index++;
        }

        return arr;
    }
}



